import re
import sys
import time
import os
import stat
import xmlrpclib
import argparse
import supervisor.xmlrpc

from supervisor.childutils import EventListenerProtocol

def get_unix_sockets(pid):
    """
    iterates over file descriptors for given pid looking for open unix sockets, yields their path names
    """
    path_by_inode = {}
    # reading open unix sockets info from /proc/net/unix
    # sample file contents:
    # ffff8800e800b800: 00000002 00000000 00010000 0001 01  9386 /var/run/supervisor.sock.1156
    # ffff8800e92cf380: 00000002 00000000 00010000 0001 01  8898 /var/run/acpid.socket
    # ffff8800e8464980: 00000002 00000000 00010000 0001 01 10193 /var/run/salt/minion/minion_event_401f1c2ee4_pub.ipc
    # ffff8800e800db00: 00000006 00000000 00000000 0002 01  8657 /dev/log
    # ffff8800e8462680: 00000002 00000000 00010000 0001 01 10195 /var/run/salt/minion/minion_event_401f1c2ee4_pull.ipc
    # ffff8800e9a4c980: 00000003 00000000 00000000 0001 03  9282
    # ffff8800e800e200: 00000003 00000000 00000000 0001 03  8103 @/com/ubuntu/upstart
    for line in open("/proc/net/unix", "r"):
        _, inode, path = line.rsplit(" ", 2)
        path_by_inode[inode] = path.strip("\n")

    for fd_num in os.listdir("/proc/{0}/fd".format(pid)):
        statinfo = os.stat('/proc/{0}/fd/{1}'.format(pid, fd_num))
        if stat.S_ISSOCK(statinfo.st_mode):
            path = path_by_inode.get(str(statinfo.st_ino))
            if not path is None: 
                if os.path.exists(path):
                    yield path
                else:
                    # stripping dot-number appendix that may be attached to the end of the file path by the os
                    # eg /var/run/supervisor.sock.1156
                    path, _ = path.rsplit(".",  1)
                    if os.path.exists(path):
                        yield path

def main():
    # supervisor_cgroups_listen [PROCESS@HIERARCHY:GROUP ...]
    # eg `supervisor_cgroups_listen kafka cpu_and_mem blue`
    # does not support cgclassify interface (user specifies subsystems such as cpu or memory instead of hierarchy) 
    # https://access.redhat.com/documentation/en-US/Red_Hat_Enterprise_Linux/6/html/Resource_Management_Guide/sec-Moving_a_Process_to_a_Control_Group.html
    parser = argparse.ArgumentParser('supervisor_cgroups_listen',
     description='Attaches supervisor processes to cgroups.\
      Groups/hierarchies must exist beforehand; you may want to edit /etc/cgconfig.conf\
       to configure them.')
    parser.add_argument("rules", metavar="PROCESSNAME@HIERARCHY:GROUP", nargs='+', help='eg kafka@mem_and_cpu:red')
    parser.add_argument("--prefix", help='cgroups prefix', default="/cgroup")
    args = parser.parse_args()
    
    possible_rpc_endpoints = [path for path in get_unix_sockets(os.getppid())]
    if not len(possible_rpc_endpoints) == 1:
        raise RuntimeError("Could not reliably determine supervisor rpc endpoint")
    rpc_endpoint = "unix://{0}".format(possible_rpc_endpoints[0])
    rpc = xmlrpclib.ServerProxy('http://127.0.0.1',
     transport=supervisor.xmlrpc.SupervisorTransport(None, None, serverurl=rpc_endpoint))

    rec = re.compile("^(.+?)@(.+?):(.+?)$", re.I|re.U)

    try:
        rules = dict([(groups[0], groups[1:]) for groups in [rec.match(rule).groups() for rule in args.rules]])
    except AttributeError:      # "AttributeError: 'NoneType' object has no attribute 'groups'"
        raise RuntimeError("Invalid rules format: `{0}`".format(" ".join(args.rules)))

    proto = EventListenerProtocol()

    def process_event(process_name):
        processinfo = rpc.supervisor.getProcessInfo(process_name)
        fd = open("{0}/{1}/{2}/tasks".format(args.prefix, *rules[process_name]), "a")
        fd.write(str(processinfo["pid"]))
        fd.close()

    while True:
        metadata, raw_program_state = proto.wait()
        if metadata["eventname"] == "SUPERVISOR_STATE_CHANGE_RUNNING":
            for process_name in rules.keys():
                process_event(process_name)
        elif metadata["eventname"] == "PROCESS_STATE_STARTING":
            program_state = dict([a.split(":") for a in raw_program_state.split(" ")])
            if not program_state["processname"] in rules.keys():
                proto.ok()
                continue
            process_event(program_state["processname"])
        proto.ok()

if __name__ == "__main__":
    main()